/*
 * Distributed under the terms of the GNU General Public License v2
 * $Header: /var/cvsroot/gentoo-projects/hardened/policycoreutils-extra/src/runscript_selinux.c,v 1.15 2005/02/23 02:40:19 pebenito Exp $
 */

/*
 * This is a modified NSA run_init for use in Gentoo's /sbin/runscript.
 */

/************************************************************************
 * BUILD OPTIONS:
 *
 * option USE_PAM:
 *
 * Set the USE_PAM constant if you want to authenticate users via PAM.
 * If USE_PAM is not set, users will be authenticated via direct
 * access to the shadow password file.
 *
 * If you decide to use PAM must be told how to handle run_init.  A
 * good rule-of-thumb might be to tell PAM to handle run_init in the
 * same way it handles su, except that you should remove the pam_rootok.so
 * entry so that even root must re-authenticate to run the init scripts
 * in the proper context.
 *
 * If you choose not to use PAM, make sure you have a shadow passwd file
 * in /etc/shadow.  You can use a symlink if your shadow passwd file
 * lives in another directory.  Example:
 *   su
 *   cd /etc
 *   ln -s /etc/auth/shadow shadow
 *
 * If you decide not to use PAM, you will also have to make run_init
 * setuid root, so that it can read the shadow passwd file.
 *************************************************************************/

#include <stdio.h>
#include <stdlib.h>               /* for malloc(), realloc(), free() */
#include <pwd.h>                  /* for getpwuid() */
#include <sys/types.h>            /* to make getuid() and getpwuid() happy */
#include <sys/wait.h>		  /* for wait() */
#include <sys/stat.h>		  /* for struct stat and friends */
#include <getopt.h>               /* for getopt_long() form of getopt() */
#include <selinux/selinux.h>
#include <selinux/get_default_type.h>
#include <selinux/context.h>              /* for context-mangling functions */
#include <fcntl.h>
#include <ctype.h>
#include <limits.h>
#include <pty.h>
#include <errno.h>
#include <signal.h>
#include <unistd.h>

#define SELECT_TIMEOUT 100000

#define RUN_INIT_FILE "run_init_type"
#define INITRC_FILE "initrc_context"
#ifdef USE_PAM

#undef MAX
#define MAX(x,y) ((x) > (y) ? (x) : (y))

/************************************************************************
 *
 * All PAM code goes in this section.
 *
 ************************************************************************/

#include <unistd.h>               /* for getuid(), exit(), getopt() */

#include <security/pam_appl.h>    /* for PAM functions */
#include <security/pam_misc.h>    /* for misc_conv PAM utility function */

#define SERVICE_NAME "run_init"   /* the name of this program for PAM */
				  /* The file containing the context to run 
				   * the scripts under.                     */

int authenticate_via_pam( const struct passwd * );

/* authenticate_via_pam()
 *
 * in:     p_passwd_line - struct containing data from our user's line in 
 *                         the passwd file.
 * out:    nothing
 * return: value   condition
 *         -----   ---------
 *           1     PAM thinks that the user authenticated themselves properly
 *           0     otherwise
 *
 * This function uses PAM to authenticate the user running this
 * program.  This is the only function in this program that makes PAM
 * calls.
 *
 */

int authenticate_via_pam( const struct passwd *p_passwd_line ) {

  int result = 0;    /* our result, set to 0 (not authenticated) by default */
  pam_handle_t *pam_handle;      /* opaque handle used by all PAM functions */

  /* This is a jump table of functions for PAM to use when it wants to *
   * communicate with the user.  We'll be using misc_conv(), which is  *
   * provided for us via pam_misc.h.                                   */
  struct pam_conv pam_conversation = {
    misc_conv,
    NULL
  };

  /* Make `p_pam_handle' a valid PAM handle so we can use it when *
   * calling PAM functions.                                       */
  if( PAM_SUCCESS != pam_start( SERVICE_NAME,
				p_passwd_line->pw_name,
				&pam_conversation,
				&pam_handle ) ) {
    fprintf( stderr, "failed to initialize PAM\n" );
    exit( -1 );
  }

  /* Ask PAM to authenticate the user running this program */
  if( PAM_SUCCESS == pam_authenticate(pam_handle,0) ) {
    result = 1;  /* user authenticated OK! */
  }

  /* We're done with PAM.  Free `pam_handle'. */
  pam_end( pam_handle, PAM_SUCCESS );
 
  return( result );

} /* authenticate_via_pam() */

#else /* else !USE_PAM */


/************************************************************************
 *
 * All shadow passwd code goes in this section.
 *
 ************************************************************************/

#include <unistd.h>                         /* for getuid(), exit(), crypt() */
#include <crypt.h>
#include <shadow.h>                         /* for shadow passwd functions */
#include <string.h>                         /* for strlen(), memset() */

#define PASSWORD_PROMPT "Password:"         /* prompt for getpass() */

int authenticate_via_shadow_passwd( const struct passwd * );

/* authenticate_via_shadow_passwd()
 *
 * in:     p_passwd_line - struct containing data from our user's line in 
 *                         the passwd file.
 * out:    nothing
 * return: value   condition
 *         -----   ---------
 *           1     user authenticated themselves properly according to the
 *                 shadow passwd file.
 *           0     otherwise
 *
 * This function uses the shadow passwd file to authenticate the user running
 * this program.
 *
 */

int authenticate_via_shadow_passwd( const struct passwd *p_passwd_line ) {

  struct spwd *p_shadow_line; /* struct derived from shadow passwd file line */
  char *unencrypted_password_s;        /* unencrypted password input by user */
  char *encrypted_password_s; /* user's password input after being crypt()ed */

  /* Make `p_shadow_line' point to the data from the current user's *
   * line in the shadow passwd file.                                */
  setspent();            /* Begin access to the shadow passwd file. */
  p_shadow_line = getspnam( p_passwd_line->pw_name );
  endspent();            /* End access to the shadow passwd file. */
  if( !( p_shadow_line ) ) {
    fprintf( stderr, "Cannot find your entry in the shadow passwd file.\n" );
    exit( -1 );
  }

  /* Ask user to input unencrypted password */
  if( ! ( unencrypted_password_s = getpass( PASSWORD_PROMPT ) ) ) {
    fprintf( stderr, "getpass cannot open /dev/tty\n" );
    exit( -1 );
  }

  /* Use crypt() to encrypt user's input password.  Clear the *
   * unencrypted password as soon as we're done, so it is not * 
   * visible to memory snoopers.                              */
  encrypted_password_s = crypt( unencrypted_password_s,
				p_shadow_line->sp_pwdp );
  memset( unencrypted_password_s, 0, strlen( unencrypted_password_s ) );

  /* Return 1 (authenticated) iff the encrypted version of the user's *
   * input password matches the encrypted password stored in the      *
   * shadow password file.                                            */
  return( !strcmp( encrypted_password_s, p_shadow_line->sp_pwdp ) );

} /* authenticate_via_shadow_passwd() */

#endif /* if/else USE_PAM */

/*
 * authenticate_user()
 *
 * Authenticate the user.
 *
 * in:		nothing
 * out:		nothing
 * return:	0 When success
 *		-1 When failure
 */
int authenticate_user() {

#define INITLEN 255
  security_context_t old_context;     /* original security context as a string */
  context_t ocontext;                 /* manipulatable form of context_s */

  struct passwd *p_passwd_line;      /* struct derived from passwd file line */

  /* Put our osid into `osid'. */
  if( 0!=(getprevcon(&old_context)) ) {
    fprintf(stderr,"failed to get old_context.\n");
    exit(-1);
  }

  ocontext=context_new(old_context);                                
  freecon(old_context);
  if( !(p_passwd_line=getpwnam(context_user_get(ocontext))) ) {
    fprintf(stderr, "cannot find your entry in the passwd file.\n");
    return(-1);
  }
  context_free( ocontext );
  
  printf("Authenticating %s.\n",p_passwd_line->pw_name);

  /* Authenticate the user running this program. */
#ifdef USE_PAM
  if( !authenticate_via_pam(p_passwd_line) ) {
#else /* !USE_PAM */
  if( !authenticate_via_shadow_passwd(p_passwd_line) ) {
#endif /* if/else USE_PAM */
    fprintf(stderr, "runscript: incorrect password for %s\n", p_passwd_line->pw_name);
    return(-1);
  }

  /* If we reach here, then we have authenticated the user. */
#ifdef CANTSPELLGDB
  printf( "You are authenticated!\n" );
#endif  

  return 0;

} /* authenticate_user() */

/*
 * get_init_context()
 *
 * Get the CONTEXT associated with the context for the init scripts.             *
 *
 * in:		nothing
 * out:		The CONTEXT associated with the context.
 * return:	0 on success, -1 on failure.
 */
int get_context_from_file(const char *filename, security_context_t *context) {

  FILE* fp;
  char  buf[255], *bufp;
  int buf_len;
  char context_file[PATH_MAX];

  memset(context_file, '\0', sizeof(context_file));
  snprintf(context_file, sizeof(context_file)-1, "%s/%s", selinux_contexts_path(), filename);
  fp = fopen(context_file, "r");
  if (!fp) {
    fprintf(stderr, "runscript_selinux.c: Could not open file %s\n", context_file);
    return -1;
  }

  while (1) {           /* loop until we find a non-empty line */

    if (!fgets(buf, sizeof buf, fp))
      break;

    buf_len = strlen(buf);
    if (buf[buf_len-1] == '\n')
      buf[buf_len-1] = 0;

    bufp = buf;
    while (*bufp && isspace(*bufp))
      bufp++;

    if( *bufp ) {
      *context = strdup(bufp);
      if (!(*context))
	goto out;
      fclose(fp);
      return 0;
    }
  }
out:
  fclose(fp);
  fprintf(stderr, "runscript_selinux.c: No context in file %s\n", context_file);
  return -1;

} /* get_context_from_file() */

int chld_received=0;

static void signal_handler(int signo) {

	switch( signo ) {
		case SIGCHLD:
			chld_received=1;
			break;
		default:		
			fprintf(stderr,"Bad signal number received by signal_handler: %d\n",signo);
			break;
	}
}

/*****************************************************************************
 * run_init()                                                                *
 *****************************************************************************/
void run_init(int argc, char **argv) {

  security_context_t new_context;    /* context for the init script context  */
  security_context_t curr_context;   /* context we are currently in */

  /* use this to extract the type */
  context_t curr_con;
  const char *curr_type, *run_init_type;

  /*
   * Return, if selinux is disabled.
   */
  if( is_selinux_enabled() < 1 ) {
#ifdef CANTSPELLGDB
    printf("SELinux is disabled.  Skipping selinux_runscript().\n");
#endif
    return;
  }
  
  /*
   * Get the run_init type (knowing that security_context_t is char*)
   */
  if( get_context_from_file(RUN_INIT_FILE,(security_context_t*)&run_init_type) == 0 ) {
#ifdef CANTSPELLGDB
    printf("run_init type is %s\n", run_init_type);
#endif
  } else {
    /* assume a reasonable default, rather than bailing out */
    run_init_type="run_init_t";
    fprintf(stderr, "runscript_selinux.c: Assuming run_init type is %s\n", run_init_type);
  }

 /*
  * Get our current context.
  */
  if ( getcon(&curr_context) == 0 ) {
#ifdef CANTSPELLGDB
    printf("current context is %s\n", curr_context);
#endif
  } else {
    exit(1);
  }

  /* extract the type from the context */
  curr_con=context_new(curr_context);
  curr_type=context_type_get(curr_con);

 /*
  * Authenticate if we're in the run_init domain.  If we are not
  * in the run_init domain, then either we are already in initrc_t
  * or things are mislabeled, and this won't work right anyway.
  */
  if( strncmp(run_init_type,curr_type,strlen(run_init_type)) == 0 ) {
    if( authenticate_user() != 0 ) {
      fprintf(stderr, "Authentication failed.\n");
      exit(1);
    }

  /*
   * Get the context for the script to be run in.
   */
    if( get_context_from_file(INITRC_FILE,&new_context) == 0 ) { 
#ifdef CANTSPELLGDB
      printf("initrc context is %s\n", new_context);
#endif  
    } else {
      /* assume a reasonable default, rather than bailing out */
      new_context="system_u:system_r:initrc_t";
      fprintf(stderr, "runscript_selinux.c: Assuming initrc context is %s\n", new_context);
    }

   /*
    * Set the correct context since we're not in the right domain.
    */

    if (setexeccon(new_context) < 0) {
       fprintf(stderr, "Could not set exec context to %s.\n", new_context);
       exit(1);
    }

    freecon(new_context);

    if( argv != NULL ) {
      /* prepend "open_init_pty" to args */
      argv[0]="open_init_pty";

      /* reexec for open_init_pty */
      if( execv("/sbin/runscript",argv) < 0 ) {
         perror("open_init_pty execv");
         exit(1);
      }
    }
  }

  context_free(curr_con);
  freecon(curr_context);

} /* run_init() */

void open_init_pty(int argc, char **argv) {

	char buf[16384];
	int rc, size, init_pty, admin_in=fileno(stdin), admin_out=fileno(stdout);
	fd_set select_fds;

	struct winsize ttysize;
	struct termios ttyattr;

	struct timeval tv;

	/* set up SIGCHLD */
	if( signal(SIGCHLD, signal_handler) == SIG_ERR ) {
		printf("Unable to set parent SIGCHLD handling\n");
		exit(1);
	}

	/* get terminal attribs */
	if( ioctl(fileno(stdout), TCGETS, &ttyattr) < 0 ) {
		perror("ioctl stdout TCGETS");
		exit(1);
	}

	/* get window size */
	if( ioctl(fileno(stdout), TIOCGWINSZ, &ttysize) < 0 ) {
		perror("ioctl stdout TIOCGWINSZ");
		exit(1);
	}

	/* forkpty() */
	rc=forkpty(&init_pty, NULL, &ttyattr, &ttysize);
	if( rc == 0 ) {
		/* child process */

		/* reset SIGCHLD to default */
		if( signal(SIGCHLD, SIG_DFL) == SIG_ERR ) {
			printf("Unable to reset child SIGCHLD handling\n");
			exit(1);
		}

		/* strip "open_init_pty" from args */
		argv[0]="/sbin/runscript.sh";

		/* if child: execv */
		if( execv("/sbin/runscript.sh",argv) < 0 ) {
			perror("runscript.sh execv");
			exit(1);
		}
	} else if( rc < 0 ) {
		perror("forkpty");
		exit(1);
	}

	/* child doesnt get here */

	/* get init pty flags and turn on nonblocking i/o */
	rc=fcntl(init_pty,F_GETFL);
	if(rc < 0) {
		perror("fcntl(init_pty,F_GETFL)");
		exit(1);
	}
	if(fcntl(init_pty,F_SETFL, rc | O_NONBLOCK) < 0) {
		perror("fcntl(init_pty,F_SETFL,rc|O_NONBLOCK)");
		exit(1);
	}

	/* get admin_in term flags and turn on nonblocking i/o */
	rc=fcntl(admin_in,F_GETFL);
	if(rc < 0) {
		perror("fcntl(admin_in,F_GETFL)");
		exit(1);
	}
	if(fcntl(admin_in,F_SETFL, rc | O_NONBLOCK) < 0) {
		perror("fcntl(admin_in,F_SETFL,rc|O_NONBLOCK)");
		exit(1);
	}

	/* while no signal */
	while(chld_received == 0) {
		/* this is a potential write after read race condition
		 * since there's no mutex protecting chld_received, but
		 * at worst we do an extra iteration, wasting around
		 * SELECT_TIMEOUT microseconds.
		 */

		/* have to reset select_fds and tv each iteration
		 * as select alters them.
		 */
		FD_ZERO(&select_fds);
		FD_SET(admin_in, &select_fds);
		FD_SET(init_pty, &select_fds);
		tv.tv_sec=0;
		tv.tv_usec=SELECT_TIMEOUT;

		rc=select(MAX(admin_in,init_pty)+1,&select_fds,NULL,NULL,&tv);
			
		switch( rc ) {
			case -1:
				perror("select");
				exit(1);
				break;
			case 0:
				continue;
				break;
			default:
				if( FD_ISSET(init_pty, &select_fds) ) {
					/* non-blocking read the pty */
					while( (size=read(init_pty, buf, 16384)) > 0 )
						/* if something is read, write it out */
						write(fileno(stdout), buf, size);
				}
				if( FD_ISSET(admin_in, &select_fds) ) {
					/* non-blocking read stdin */
					while( (size=read(admin_in, buf, 16384)) > 0 )
						/* if something is read, write it out */
						write(init_pty, buf, size);
				}
				break;
		}
	}

	/* leftover reads from init pty */
	while( (size=read(init_pty, buf, 16384)) > 0 )
		/* if something is read, write it out */
		write(admin_out, buf, size);

	/* don't bother reading anymore from admin_in,
	 * since the child has already terminated.
	 */

	fflush(stdout);
	exit(0);
}

void selinux_runscript() {
	/* legacy support */
#ifdef CANTSPELLGDB
	printf("Using legacy run_init code.\n");
#endif
	run_init(0,NULL);
}

void selinux_runscript2(int argc, char **argv) {

	/* kind of a hack.  if we're doing a run_init,
	 * do the setexec, and then re-exec /sbin/runscript
	 * with "open_init_pty" as argv[0].
	 *
	 * on the second run through, strip the dummy
	 * open_init_pty arg, forkpty, then exec the
	 * real runscript
	 */

	if( strncmp(argv[0],"open_init_pty",strlen("open_init_pty")+1) == 0) {
#ifdef CANTSPELLGDB
		printf("Using open_init_pty code.\n");
#endif
		/* open_init_pty */
		open_init_pty(argc,argv);
		/* shouldn't get this far */
		exit(127);
	} else {
#ifdef CANTSPELLGDB
		printf("Using new run_init code.\n");
#endif
		/* run_init */
		run_init(argc,argv);
		/* should only get this far if selinux
		 * is disabled or there is no setexec
		 */
	}
}
